@@ -67,6 +67,7 @@ epoch_period = 1e30
 last_ax = None # always assume we want to plot in the last axis, unless explicitly specified
 overlay_axs = [] # for keeping track of candlesticks in overlays
 viewrestore = False
+master_data = {}
 
 
 
@@ -83,7 +84,7 @@ class EpochAxisItem(pg.AxisItem):
         if self.mode == 'num':
             return ['%g'%v for v in values]
         conv = _x2year if self.mode=='year' else _x2local_t
-        return [conv(self.vb.datasrc, value) for value in values]
+        return [conv(self.vb.datasrc, value)[0] for value in values]
 
     def tickValues(self, minVal, maxVal, size):
         self.mode = 'num'
@@ -99,8 +100,8 @@ class EpochAxisItem(pg.AxisItem):
         # year index calculation
         self.mode = 'year'
         maxVal = min(datasrc.df.index[-1], maxVal)
-        y0 = int(_x2utc(datasrc, minVal)[:4])
-        y1 = int(_x2utc(datasrc, maxVal)[:4])
+        y0 = int(_x2utc(datasrc, minVal)[0][:4])
+        y1 = int(_x2utc(datasrc, maxVal)[0][:4])
         step = (y1-y0)//12 or 1
         years = pd.Series(pd.to_datetime(['%s'%y for y in range(y0,y1+1,step)]))
         years_indices = [ceil(yi) for yi in _pdtime2index(ax, years)]
@@ -281,6 +282,11 @@ class PandasDataSource:
         self.cache_hilo = OrderedDict()
         self._period = None
 
+    def set_df(self, df):
+        self.df = df
+        self.cache_hilo = OrderedDict()
+        self._period = None
+
     def hilo(self, x0, x1):
         '''Return five values in time range: t0, t1, highest, lowest, number of rows.'''
         if x0 == x1:
@@ -358,8 +364,13 @@ class FinWindow(pg.GraphicsLayoutWidget):
         self.setGeometry(winx, winy, winw, winh)
         winx += 40
         winy += 40
-        self.show()
         self.centralWidget.installEventFilter(self)
+        self.ci.setContentsMargins(0, 0, 0, 0)
+        self.ci.setSpacing(-1)
+
+    @property
+    def axs(self):
+        return [ax for ax in self.ci.items if isinstance(ax, pg.PlotItem)]
 
     def close(self):
         _savewindata(self)
@@ -409,13 +420,17 @@ class FinCrossHair:
         rng = self.ax.vb.y_max - self.ax.vb.y_min
         rngmax = abs(self.ax.vb.y_min) + rng # any approximation is fine
         sd,se = (self.ax.significant_decimals,self.ax.significant_eps) if clamp_grid else (significant_decimals,significant_eps)
+        timebased = False
         if self.ax.x_indexed:
-            xtext = _x2local_t(self.ax.vb.datasrc, x)
+            xtext,timebased = _x2local_t(self.ax.vb.datasrc, x)
         else:
             xtext = _round_to_significant(rng, rngmax, x, sd, se)
         linear_y = y
         y = self.ax.vb.yscale.xform(y)
         ytext = _round_to_significant(rng, rngmax, y, sd, se)
+        if not timebased:
+            xtext = 'x ' + xtext
+            ytext = 'y ' + ytext
         far_right = self.ax.viewRect().x() + self.ax.viewRect().width()*0.9
         far_bottom = self.ax.viewRect().y() + self.ax.viewRect().height()*0.1
         close2right = x > far_right
@@ -563,6 +578,7 @@ class FinViewBox(pg.ViewBox):
         self.set_datasrc(None)
         self.setMouseEnabled(x=True, y=False)
         self.init_steps = init_steps
+        self.updating_linked = False
 
     def set_datasrc(self, datasrc):
         self.datasrc = datasrc
@@ -671,9 +687,10 @@ class FinViewBox(pg.ViewBox):
         super().keyPressEvent(ev)
 
     def linkedViewChanged(self, view, axis):
-        if not self.datasrc:
+        if not self.datasrc or self.updating_linked:
             return
         if view and self.datasrc and view.datasrc:
+            self.updating_linked = True
             tr = self.targetRect()
             vr = view.targetRect()
             is_dirty = view.force_range_update > 0
@@ -692,6 +709,7 @@ class FinViewBox(pg.ViewBox):
                         view.force_range_update -= 1
                     x0,x1 = _pdtime2index(self.parent(), pd.Series([vt0,vt1]), any_end=True)
                     self.update_y_zoom(x0, x1)
+            self.updating_linked = False
 
     def zoom_rect(self, vr, scale_fact, center):
         if not self.datasrc:
@@ -722,8 +740,8 @@ class FinViewBox(pg.ViewBox):
         main_vb = self
         if self.linkedView(0):
             self.force_range_update = 1 # main need to update only once to us
-            main_vb = list(self.win.ci.items)[0].vb
-        main_vb.force_range_update = len(self.win.ci.items)-1 # update main as many times as there are other rows
+            main_vb = list(self.win.axs)[0].vb
+        main_vb.force_range_update = len(self.win.axs)-1 # update main as many times as there are other rows
         self.update_y_zoom()
         # refresh crosshair when done
         _mouse_moved(self.win, None)
@@ -749,7 +767,7 @@ class FinViewBox(pg.ViewBox):
         # fetch hi-lo and set range
         _,_,hi,lo,cnt = self.datasrc.hilo(x0, x1)
         vr = self.viewRect()
-        if cnt < vr.width() and cnt < max_zoom_points:
+        if (x1-x0) < vr.width() and cnt < max_zoom_points:
             return
         if not self.v_autozoom:
             hi = vr.bottom()
@@ -1083,37 +1101,48 @@ class ScatterLabelItem(FinPlotItem):
         return self.viewRect()
 
 
-
 def create_plot(title='Finance Plot', rows=1, init_zoom_periods=1e10, maximize=True, yscale='linear'):
-    global windows, last_ax
     pg.setConfigOptions(foreground=foreground, background=background)
     win = FinWindow(title)
-    windows.append(win)
-    if maximize:
-        win.showMaximized()
     # normally first graph is of higher significance, so enlarge
     win.ci.layout.setRowStretchFactor(0, top_graph_scale)
-    win.ci.setContentsMargins(0, 0, 0, 0)
-    win.ci.setSpacing(-1)
+    if maximize:
+        win.showMaximized()
+    ax0 = axs = create_plot_widget(master=win, rows=rows, init_zoom_periods=init_zoom_periods, yscale=yscale)
+    axs = axs if type(axs) in (tuple,list) else [axs]
+    for ax in axs:
+        win.addItem(ax)
+        win.nextRow()
+    return ax0
+
+
+def create_plot_widget(master, rows=1, init_zoom_periods=1e10, yscale='linear'):
+    pg.setConfigOptions(foreground=foreground, background=background)
+    global last_ax
+    if master not in windows:
+        windows.append(master)
     axs = []
     prev_ax = None
     for n in range(rows):
         ysc = yscale[n] if type(yscale) in (list,tuple) else yscale
         ysc = YScale(ysc, 1)
         v_zoom_scale = 0.97
-        viewbox = FinViewBox(win, init_steps=init_zoom_periods, yscale=ysc, v_zoom_scale=v_zoom_scale)
-        ax = prev_ax = _add_timestamp_plot(win, prev_ax, viewbox=viewbox, index=n, yscale=ysc)
-        _set_plot_x_axis_leader(ax)
-        if n == 0:
+        viewbox = FinViewBox(master, init_steps=init_zoom_periods, yscale=ysc, v_zoom_scale=v_zoom_scale)
+        ax = prev_ax = _add_timestamp_plot(master=master, prev_ax=prev_ax, viewbox=viewbox, index=n, yscale=ysc)
+        if axs:
+            ax.setXLink(axs[0].vb)
+        else:
             viewbox.setFocus()
         axs += [ax]
-    win.proxy_mmove = pg.SignalProxy(win.scene().sigMouseMoved, rateLimit=144, slot=partial(_mouse_moved, win))
-    win._last_mouse_evs = None
-    win._last_mouse_y = 0
+    if isinstance(master, FinWindow):
+        proxy = pg.SignalProxy(master.scene().sigMouseMoved, rateLimit=144, slot=partial(_mouse_moved, master))
+    else:
+        proxy = []
+        for ax in axs:
+            proxy += [pg.SignalProxy(ax.ax_widget.scene().sigMouseMoved, rateLimit=144, slot=partial(_mouse_moved, master))]
+    master_data[master] = dict(proxymm=proxy, last_mouse_evs=None, last_mouse_y=0)
     last_ax = axs[0]
-    if len(axs) == 1:
-        return axs[0]
-    return axs
+    return axs[0] if len(axs) == 1 else axs
 
 
 def price_colorfilter(item, datasrc, df):
@@ -1145,6 +1174,17 @@ def strength_colorfilter(item, datasrc, df):
     yield item.rowcolors('bear') + [df.loc[(~is_up)&(~is_strong), :]]
 
 
+def volume_colorfilter_section(sections=[]):
+    '''The sections argument is a (starting_index, color_name) array.'''
+    def _colorfilter(sections, item, datasrc, df):
+        if not sections:
+            return volume_colorfilter(item, datasrc, df)
+        for (i0,colname),(i1,_) in zip(sections, sections[1:]+[(None,'neutral')]):
+            rows = df.iloc[i0:i1, :]
+            yield item.rowcolors(colname) + [rows]
+    return partial(_colorfilter, sections)
+
+
 def horizvol_colorfilter(sections=[]):
     '''The sections argument is a (starting_index, color_name) array.'''
     def _colorfilter(sections, item, datasrc, data):
@@ -1378,7 +1418,7 @@ def set_y_range(ymin, ymax, ax=None):
     ax.vb.set_range(None, ymin, None, ymax)
 
 
-def set_yscale(yscale='linear', ax=None):
+def set_y_scale(yscale='linear', ax=None):
     ax = _create_plot(ax=ax, maximize=False)
     ax.setLogMode(y=(yscale=='log'))
     ax.vb.yscale = YScale(yscale, ax.vb.yscale.scalef)
@@ -1471,9 +1511,9 @@ def autoviewrestore(enable=True):
     viewrestore = enable
 
 
-def show():
+def show(qt_exec=True):
     for win in windows:
-        vbs = [ax.vb for ax in win.ci.items]
+        vbs = [ax.vb for ax in win.axs]
         for vb in vbs:
             _pre_process_data(vb)
         if viewrestore:
@@ -1483,12 +1523,18 @@ def show():
             if vb.datasrc and (vb.linkedView(0) is None or vb.linkedView(0).datasrc is None):
                 vb.update_y_zoom(vb.datasrc.init_x0, vb.datasrc.init_x1)
     _repaint_candles()
-    if windows:
+    for win in windows:
+        if isinstance(win, FinWindow) or qt_exec:
+            win.show()
+    if windows and qt_exec:
         global last_ax, app
         app = QtGui.QApplication.instance()
         app.exec_()
         windows.clear()
         overlay_axs.clear()
+        timers.clear()
+        sounds.clear()
+        master_data.clear()
         last_ax = None
 
 
@@ -1501,15 +1547,17 @@ def play_sound(filename):
 
 
 def screenshot(file, fmt='png'):
-    if not app:
+    if _internal_windows_only() and not app:
         print('ERROR: save_screenshot must be callbacked from e.g. timer_callback()')
-        return
+        return False
     try:
         buffer = QtCore.QBuffer()
         app.primaryScreen().grabWindow(windows[0].winId()).save(buffer, fmt)
         file.write(buffer.data())
+        return True
     except Exception as e:
         print(type(e), e)
+    return False
 
 
 #################### INTERNALS ####################
@@ -1524,7 +1572,7 @@ def _loadwindata(win):
     except:
         return
     kvs = {k:v for k,v in settings}
-    vbs = set(ax.vb for ax in win.ci.items)
+    vbs = set(ax.vb for ax in win.axs)
     for vb in vbs:
         ds = vb.datasrc
         if ds:
@@ -1543,7 +1591,7 @@ def _savewindata(win):
     try:
         min_x = int(1e100)
         max_x = int(-1e100)
-        for ax in win.ci.items:
+        for ax in win.axs:
             if ax.vb.targetRect().right() < 4: # ignore empty plots
                 continue
             t0,t1,_,_,_ = ax.vb.datasrc.hilo(ax.vb.targetRect().left(), ax.vb.targetRect().right())
@@ -1561,6 +1609,10 @@ def _savewindata(win):
         print('Error saving plot:', e)
 
 
+def _internal_windows_only():
+     return all(isinstance(win,FinWindow) for win in windows)
+
+
 def _create_plot(ax=None, **kwargs):
     if ax:
         return ax
@@ -1569,13 +1621,18 @@ def _create_plot(ax=None, **kwargs):
     return create_plot(**kwargs)
 
 
-def _add_timestamp_plot(win, prev_ax, viewbox, index, yscale):
-    if prev_ax is not None:
+def _add_timestamp_plot(master, prev_ax, viewbox, index, yscale):
+    native_win = isinstance(master, FinWindow)
+    if native_win and prev_ax is not None:
         prev_ax.set_visible(xaxis=False) # hide the whole previous axis
-        win.nextRow()
     axes = {'bottom': EpochAxisItem(vb=viewbox, orientation='bottom'),
             'left':   YAxisItem(vb=viewbox, orientation='left')}
-    ax = pg.PlotItem(viewBox=viewbox, axisItems=axes, name='plot-%i'%index)
+    if native_win:
+        ax = pg.PlotItem(viewBox=viewbox, axisItems=axes, name='plot-%i'%index)
+    else:
+        axw = pg.PlotWidget(viewBox=viewbox, axisItems=axes, name='plot-%i'%index)
+        ax = axw.plotItem
+        ax.ax_widget = axw
     ax.axes['left']['item'].textWidth = y_label_width # this is to put all graphs on equal footing when texts vary from 0.4 to 2000000
     ax.axes['left']['item'].setStyle(tickLength=-5) # some bug, totally unexplicable (why setting the default value again would fix repaint width as axis scale down)
     ax.axes['left']['item'].setZValue(30) # put axis in front instead of behind data
@@ -1594,7 +1651,6 @@ def _add_timestamp_plot(win, prev_ax, viewbox, index, yscale):
     if index%2:
         viewbox.setBackgroundColor(odd_plot_background)
     viewbox.setParent(ax)
-    win.addItem(ax)
     return ax
 
 
@@ -1602,7 +1658,10 @@ def _overlay(ax, scale=0.25):
     global overlay_axs
     viewbox = FinViewBox(ax.vb.win, init_steps=ax.vb.init_steps, yscale=YScale('linear', 1))
     viewbox.v_zoom_scale = scale
-    ax.vb.win.centralWidget.scene().addItem(viewbox)
+    if hasattr(ax, 'ax_widget'):
+        ax.ax_widget.scene().addItem(viewbox)
+    else:
+        ax.vb.win.centralWidget.scene().addItem(viewbox)
     viewbox.setXLink(ax.vb)
     def updateView():
         viewbox.setGeometry(ax.vb.sceneBoundingRect())
@@ -1618,8 +1677,10 @@ def _overlay(ax, scale=0.25):
     axo.hideAxis('bottom')
     axo.hideButtons()
     viewbox.addItem(axo)
+    viewbox.setParent(axo)
     ax.vb.sigResized.connect(updateView)
     overlay_axs.append(axo)
+    updateView()
     return axo
 
 
@@ -1691,7 +1752,7 @@ def _create_datasrc(ax, *args, datacols=1):
     # check if time column missing
     if len(datasrc.df.columns) == datacols:
         # assume time data has already been added before
-        for a in ax.vb.win.ci.items:
+        for a in ax.vb.win.axs:
             if a.vb.datasrc and len(a.vb.datasrc.df.columns) >= 2:
                 datasrc.df.columns = a.vb.datasrc.df.columns[1:len(datasrc.df.columns)+1]
                 col = a.vb.datasrc.df.columns[0]
@@ -1714,7 +1775,7 @@ def _set_datasrc(ax, datasrc):
             viewbox.datasrc.addcols(datasrc)
             for item in ax.items:
                 if isinstance(item, FinPlotItem):
-                    item.datasrc.df = viewbox.datasrc.df # every plot here now has the same time-frame
+                    item.datasrc.set_df(viewbox.datasrc.df) # every plot here now has the same time-frame
             _set_x_limits(ax, datasrc)
             viewbox.set_datasrc(viewbox.datasrc) # update zoom
     else:
@@ -1757,7 +1818,7 @@ def _adjust_renko_datasrc(step, datasrc):
                 y = x*step
                 z = list(extras.loc[i])
                 data.append([t-td, y+ds, y+step-ds, y+step, y] + z)
-    datasrc.df = pd.DataFrame(data, columns='time open close high low'.split()+list(extras.columns))
+    datasrc.set_df(pd.DataFrame(data, columns='time open close high low'.split()+list(extras.columns)))
 
 
 def _adjust_renko_log_datasrc(step, datasrc):
@@ -1771,7 +1832,7 @@ def _adjust_renko_log_datasrc(step, datasrc):
 def _adjust_volume_datasrc(datasrc):
     if len(datasrc.df.columns) <= 4:
         datasrc.df.insert(3, '_zero_', [0]*len(datasrc.df)) # base of candles is always zero
-    datasrc.df = datasrc.df.iloc[:,[0,3,4,1,2]] # re-arrange columns for rendering
+    datasrc.set_df(datasrc.df.iloc[:,[0,3,4,1,2]]) # re-arrange columns for rendering
     datasrc.scale_cols = [1, 2] # scale by both baseline and volume
 
 
@@ -1806,7 +1867,7 @@ def _adjust_bar_datasrc(datasrc, order_cols=True):
         datasrc.df.insert(1, '_open_',  [0]*len(datasrc.df)) # "open" for color
         datasrc.df.insert(2, '_close_', datasrc.df.iloc[:, 3]) # "close" (actual bar value) for color
     if order_cols:
-        datasrc.df = datasrc.df.iloc[:,[0,3,4,1,2]] # re-arrange columns for rendering
+        datasrc.set_df(datasrc.df.iloc[:,[0,3,4,1,2]]) # re-arrange columns for rendering
     datasrc.scale_cols = [1, 2] # scale by both baseline and volume
 
 
@@ -1816,24 +1877,47 @@ def _update_data(preadjustfunc, adjustfunc, item, ds):
     ds = _create_datasrc(item.ax, ds)
     if adjustfunc:
         adjustfunc(ds)
-    item.datasrc.update(ds)
     if isinstance(item, FinPlotItem):
+        item.ax.removeItem(item)
         item.dirty = True
+        item.datasrc.update(ds)
     else:
+        item.datasrc.update(ds)
         item.setData(item.datasrc.index, item.datasrc.y)
+    for i in item.ax.items:
+        if i == item or not hasattr(i, 'datasrc'):
+            continue
+        lc = len(item.datasrc.df)
+        li = len(i.datasrc.df)
+        if lc and li and max(lc,li)/min(lc,li) > 100: # TODO: should be typed instead
+            continue
+        cc = item.datasrc.df.columns
+        ci = i.datasrc.df.columns
+        c0 = [c for c in ci if c in cc]
+        c1 = [c for c in ci if not c in cc]
+        df_clipped = item.datasrc.df[c0]
+        if c1:
+            df_clipped = df_clipped.copy()
+            for c in c1:
+                df_clipped[c] = i.datasrc.df[c]
+        i.datasrc.set_df(df_clipped)
     if not item.datasrc.standalone:
         # new limits when extending/reducing amount of data
         x_min,x1 = _set_x_limits(item.ax, item.datasrc)
         # scroll all plots if we're at the far right
         tr = item.ax.vb.targetRect()
         x0 = x1 - tr.width()
-        if tr.right() >= x1 - 5 - 2*right_margin_candles:
-            item.ax.vb.update_y_zoom(x0, x1)
+        if tr.right() < x1 - 5 - 2*right_margin_candles:
+            x0 = x1 = None
+        item.ax.vb.update_y_zoom(x0, x1)
         if item.ax.axes['bottom']['item'].isVisible(): # update axes if visible
             item.ax.axes['bottom']['item'].hide()
             item.ax.axes['bottom']['item'].show()
-    for ax in item.ax.vb.win.ci.items:
+    for ax in item.ax.vb.win.axs:
         ax.vb.update()
+    if isinstance(item, FinPlotItem):
+        item.ax.addItem(item)
+        item.repaint()
 
 
 def _pre_process_data(vb):
@@ -1845,16 +1929,6 @@ def _pre_process_data(vb):
             vb.y_positive = False
 
 
-def _set_plot_x_axis_leader(ax):
-    '''The first plot to add some data is the leader. All other's X-axis will follow this one.'''
-    if ax.vb.linkedView(0):
-        return
-    for _ax in ax.vb.win.ci.items:
-        if not _ax.vb.linkedView(0) and _ax.vb.name != ax.vb.name:
-            ax.setXLink(_ax.vb.name)
-            break
-
-
 def _set_x_limits(ax, datasrc):
     if ax.x_indexed:
         x0 = -side_margin
@@ -1872,7 +1946,7 @@ def _set_x_limits(ax, datasrc):
 
 def _repaint_candles():
     '''Candles are only partially drawn, and therefore needs manual dirty reminder whenever it goes off-screen.'''
-    axs = [ax for win in windows for ax in win.ci.items] + overlay_axs
+    axs = [ax for win in windows for ax in win.axs] + overlay_axs
     for ax in axs:
         for item in ax.items:
             if isinstance(item, FinPlotItem):
@@ -1889,7 +1963,7 @@ def _key_pressed(vb, ev):
         global clamp_grid
         clamp_grid = not clamp_grid
         for win in windows:
-            for ax in win.ci.items:
+            for ax in win.axs:
                 ax.crosshair.update()
     elif ev.text() in ('\r', ' '): # enter, space
         vb.set_draw_line_color(draw_done_color)
@@ -1924,21 +1998,21 @@ def _mouse_clicked(vb, ev):
     return True
 
 
-def _mouse_moved(win, evs):
+def _mouse_moved(master, evs):
     if not evs:
-        evs = win._last_mouse_evs
+        evs = master_data[master]['last_mouse_evs']
         if not evs:
             return
-    win._last_mouse_evs = evs
+    master_data[master]['last_mouse_evs'] = evs
     pos = evs[-1]
     # allow inter-pixel moves if moving mouse slowly
     y = pos.y()
-    dy = y - win._last_mouse_y
+    dy = y - master_data[master]['last_mouse_y']
     if 0 < abs(dy) <= 1:
         pos.setY(pos.y() - dy/2)
-    win._last_mouse_y = y
+    master_data[master]['last_mouse_y'] = y
     # apply to all crosshairs
-    for ax in win.ci.items:
+    for ax in master.axs:
         point = ax.vb.mapSceneToView(pos)
         if ax.crosshair:
             ax.crosshair.update(point)
@@ -2043,7 +2117,7 @@ def _pdtime2index(ax, ts, any_end=False, require_time=False):
 def _get_datasrc(ax, require=True):
     if ax.vb.datasrc is not None or not ax.x_indexed:
         return ax.vb.datasrc
-    vbs = set(ax.vb for win in windows for ax in win.ci.items)
+    vbs = [ax.vb for win in windows for ax in win.axs]
     for vb in vbs:
         if vb.datasrc:
             return vb.datasrc
@@ -2061,13 +2135,13 @@ def _x2utc(datasrc, x):
 
 def _x2t(datasrc, x, ts2str):
     if not datasrc:
-        return ''
+        return '',False
     try:
         x += 0.5
         t,_,_,_,cnt = datasrc.hilo(x, x)
         if cnt:
             if not datasrc.timebased():
-                return '%g' % t
+                return '%g' % t, False
             s = ts2str(t)
             if epoch_period >= 24*60*60:
                 i = s.index(' ')
@@ -2079,15 +2153,16 @@ def _x2t(datasrc, x, ts2str):
                 i = -3
             else:
                 i = len(s)
-            return s[:i]
+            return s[:i],True
     except Exception as e:
         import traceback
         traceback.print_exc()
-    return ''
+    return '',datasrc.timebased()
 
 
 def _x2year(datasrc, x):
-    return _x2local_t(datasrc, x)[:4]
+    t,hasds = _x2local_t(datasrc, x)
+    return t[:4],hasds
 
 
 def _round_to_significant(rng, rngmax, x, significant_decimals, significant_eps):
